/*
 *  Copyright 2001-2010 Mikhail Kryshen <mikhail@kryshen.net>
 *
 *  This file is part of Graph Game.
 *
 *  Graph Game is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Graph Game is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Graph Game.  If not, see <http://www.gnu.org/licenses/>.
 */
package kryshen.graphg;

import java.awt.Button;
import java.awt.Checkbox;
import java.awt.CheckboxGroup;
import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.Cursor;
import java.awt.Desktop;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Frame;
import java.awt.Graphics;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Image;
import java.awt.Insets;
import java.awt.Label;
import java.awt.Panel;
import java.awt.Scrollbar;
import java.awt.TextField;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.net.URL;
import kryshen.util.ResourceLoader;

public class Game {
    public static final String TITLE = "Graph Game";
    public static final String VERSION = "1.1a";
    public static final String AUTHOR = "Mikhail Kryshen";
    public static final String HOMEPAGE = 
        "http://kryshen.pp.ru/applets/graphg.html";

    /* Layout.
     *
     * TODO: read values from skin resource.
     */
    private final int width = 494;
    private final int height = 344;
    private final int fieldX = 27;
    private final int fieldY = 70;
    private final int fieldWidth = 439;
    private final int fieldHeight = 247;
    private final int statusX = 212;
    private final int statusY = 20;
    private final int statusWidth = 254;
    private final int statusHeight = 26;
    private final int settingsBorder = 15;
    private final int urlX = 3;
    private final int urlY = 218;
    private final int urlWidth = 310;
    private final int urlHeight = 24;
    private final int buttonWidth = 88;
    private final int buttonHeight = 22;
    private final int startButtonX = 23;
    private final int startButtonY = 23;
    private final int aboutButtonX = 114;
    private final int aboutButtonY = 23;

    private final Font statusFont = new Font("Helvetica", Font.BOLD, 12);

    private final Color backgroundColor = Color.WHITE;
    private final Color settingsBackground = new Color(218, 218, 218);
    private final Color statusColor1 = Color.RED;
    private final Color statusColor2 = Color.BLUE;
    private final Color statusColor3 = Color.BLUE;

    private String status1 = TITLE + " " + VERSION;
    private String status2 = "Loading...";
    private String status3 = null;

    Container container;
    Applet applet = null;
    Frame frame = null;

    private Image about = null;
    private Image background = null;
    private Image urlBorder = null;

    private AboutPane aboutPane;
    private GraphCanvas field;
    private SettingsPane settingsPane;

    private NICButton newGameButton;
    private NICButton aboutButton;

    public static void main(String[] args) {
	final Game game = new Game();

	final Frame frame = new Frame(TITLE + " " + VERSION) {
		private RenderingManager rm = new RenderingManager();

		public void update(Graphics g) {
		    paint(g);
		}

		public void paint(Graphics g) {
		    rm.updateBuffer(this, g, getSize());
		    game.render(rm.offscreen);
		    super.paint(rm.offscreen);
		    g.drawImage(rm.buffer, 0, 0, null);
		}

		public Dimension getPreferredSize() {
		    Insets i = getInsets();
                    return new Dimension(
                            game.width + i.left + i.right,
                            game.height + i.top + i.bottom);
		}
	    };

	frame.addWindowListener(new WindowAdapter() {
		public void windowClosing(WindowEvent e) {
		    game.destroy();
		    frame.dispose();
		}
	    });
	

	frame.pack();
	frame.setResizable(false);

	Toolkit t = Toolkit.getDefaultToolkit();
	frame.setLocation((t.getScreenSize().width - frame.getSize().width) / 2,
			  (t.getScreenSize().height - frame.getSize().height) / 2);

        game.setContainer(frame);
	frame.setVisible(true);
        game.init();

	try {
	    game.load(new FileInputStream("skins/default.bin"));
	} catch (Exception e) {
	    game.displayError(e);
	}
    }

    void setContainer(Container c) {
        this.container = c;

	if (c instanceof Applet) {
	    applet = (Applet)c;
	} 

        if (c instanceof Frame) {
            frame = (Frame)c;
        }
    }

    void init() {
	container.setLayout(null);
        container.setBackground(backgroundColor);
	
	Insets insets = container.getInsets();

	aboutPane = new Game.AboutPane();
	aboutPane.setBounds(fieldX + insets.left, fieldY + insets.top,
			    fieldWidth, fieldHeight);
	aboutPane.setVisible(true);
	container.add(aboutPane);

	field = new GraphCanvas(this);
	field.setBounds(fieldX + insets.left, fieldY + insets.top,
			fieldWidth, fieldHeight);
	field.setVisible(false);
	container.add(field);
	
	settingsPane = new SettingsPane();
	settingsPane.setBounds(fieldX + insets.left + settingsBorder,
			       fieldY + insets.top + settingsBorder,
			       fieldWidth - settingsBorder * 2,
                               fieldHeight - settingsBorder * 2);
	settingsPane.setVisible(false);
	container.add(settingsPane);
	settingsPane.validate();
	
	newGameButton = new NICButton("New Game", buttonWidth, buttonHeight);
        newGameButton.setLocation(insets.left + startButtonX,
                insets.top + startButtonY);
	container.add(newGameButton);

	aboutButton = new NICButton("About", buttonWidth, buttonHeight);
	aboutButton.setLocation(insets.left + aboutButtonX,
                insets.top + aboutButtonY);
	container.add(aboutButton);
       
        ActionListener actionListener = new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                if ("New Game".equals(e.getActionCommand())) {
                    showSettings();
                } else if ("About".equals(e.getActionCommand())) {
                    showAboutPane();
                } else if ("Continue".equals(e.getActionCommand())) {
                    showField();
                }
            }
        };

	newGameButton.addActionListener(actionListener);
	aboutButton.addActionListener(actionListener);
    }

    private void showAboutPane() {
	if (aboutPane.isVisible()) return;

	if (field.isVisible())
	    aboutButton.setText("Continue");

	aboutPane.setVisible(true);
	settingsPane.setVisible(false);
	field.setVisible(false);
    }

    private void showField() {
	if (field.isVisible()) return;

	field.setVisible(true);
	settingsPane.setVisible(false);
	aboutPane.setVisible(false);
	aboutButton.setText("About");
    }

    private void showSettings() {
	if (settingsPane.isVisible()) return;

	if (field.isVisible())
	    aboutButton.setText("Continue");
	else
	    aboutButton.setText("About");

	settingsPane.setVisible(true);
	field.setVisible(false);
	aboutPane.setVisible(false);
    }

    void destroy() {
	field.stop();
	container.removeAll();
    }

    synchronized void render(Graphics g) {
        Insets insets = container.getInsets();

        if (background != null) {
            g.drawImage(background, insets.left, insets.top, container);
        } else {
            g.clearRect(insets.left, insets.top, width, height);
        }

        int sx = statusX + insets.left;
        int sy = statusY + insets.top;

        g.setFont(statusFont);
        g.setColor(statusColor1);

        RenderingManager.drawStringInRect(status1, sx, sy,
                statusWidth, statusHeight / 2,
                RenderingManager.ALIGN_CENTER, g);

        g.setColor(statusColor2);

        RenderingManager.drawStringInRect(status2, sx, sy + statusHeight / 2,
                (status3 == null) ? statusWidth : statusWidth / 2,
                statusHeight / 2, RenderingManager.ALIGN_CENTER, g);

        if (status3 != null) {
            g.setColor(statusColor3);

            RenderingManager.drawStringInRect(status3, sx + statusWidth / 2,
                    sy + statusHeight / 2,
                    statusWidth / 2, statusHeight / 2,
                    RenderingManager.ALIGN_CENTER, g);

        }
    }
    
    synchronized void setStatus1(String status1) {
	this.status1 = status1;
    }

    synchronized void setStatus2(String status2) {
	this.status2 = status2;
    }

    synchronized void setStatus3(String status3) {
	this.status3 = status3;
    }

    void repaintStatus() {
	Insets insets = container.getInsets();
	container.repaint(statusX + insets.left, statusY + insets.top, 
			  statusWidth, statusHeight);
    }

    void displayError(Throwable e) {
	System.err.println(TITLE + ": " + e);
	setStatus2(e.toString());
	setStatus3(null);
	container.repaint();
    }

    void load(InputStream in) throws IOException, InterruptedException {
	ResourceLoader loader = new ResourceLoader
	    (container, "Graph Game Skin", in);

	int progress = 0;

	setStatus2("Loading...");
	setStatus3("0%");
	container.repaint();

	while ((progress = loader.loadNext()) >= 0) {
	    //Thread.currentThread().sleep(1000);

            switch (loader.getResourceType()) {
                case ResourceLoader.DATA_IMAGE:
                    if ("background".equals(loader.getResourceName())) {
                        this.background = loader.getImage();
                    }
                    if ("about".equals(loader.getResourceName())) {
                        this.about = loader.getImage();
                    }
                    if ("url-border".equals(loader.getResourceName())) {
                        this.urlBorder = loader.getImage();
                    }
                    break;
                case ResourceLoader.DATA_FONT:
                    break;
                case ResourceLoader.DATA_COLOR:
                    break;
            }

	    setStatus2("Loading...");
	    setStatus3((int)(progress * 100f / 255) + "%");
	    container.repaint();
	}

	Thread.sleep(300);
	setStatus2("Applet loaded");
	setStatus3(null);
	container.repaint();
    }

    private void openInBrowser(final String address) {
        try {
            (new Object() {
                void browse(String address) throws Exception {
                    Desktop.getDesktop().browse(new URI(address));
                }
            }).browse(address);
        } catch (Throwable e) {
            displayError(e);
        }
    }

    private class AboutPane extends Component
	implements MouseListener, MouseMotionListener {

	private boolean displayUrlBorder = false;

	AboutPane() {
	    this.addMouseListener(this);
	    this.addMouseMotionListener(this);
	}

	public void mouseClicked(MouseEvent e) {
            if (displayUrlBorder) {
                if (applet != null) {
                    try {
                        applet.getAppletContext().
                            showDocument(new URL(HOMEPAGE));
                    } catch (Exception exc) {
                        displayError(exc);
                    }
                } else if (frame != null) {
                    openInBrowser(HOMEPAGE);
                }
            }
        }
        
        public void mouseEntered(MouseEvent e) {}
        
        public void mouseExited(MouseEvent e) {
            setDisplayUrlBorder(false);
        }
        
	public void mousePressed(MouseEvent e) {}
	public void mouseReleased(MouseEvent e) {}
        
        public void mouseMoved(MouseEvent e) {
            int x = e.getX(), y = e.getY();

            if (x >= urlX && x <= urlX + urlWidth &&
                    y >= urlY && y <= urlY + urlHeight) {
                setDisplayUrlBorder(true);
            } else {
                setDisplayUrlBorder(false);
            }
        }

	public void mouseDragged(MouseEvent e) {}
    
	public Dimension getPreferredSize() {
	    return new Dimension(fieldWidth, fieldHeight);
	}

	public void paint(Graphics g) {
	    if (about != null) 
		g.drawImage(about, 0, 0, this);
	    if (displayUrlBorder && urlBorder != null)
		g.drawImage(urlBorder, urlX, urlY, this);
	}

	public boolean isOpaque() {
	    return false;
	}

        private void setDisplayUrlBorder(boolean displayUrlBorder) {
            if (this.displayUrlBorder == displayUrlBorder) {
                return;
            }

            this.displayUrlBorder = displayUrlBorder;
            repaint();

            if (displayUrlBorder) {
                setCursor(new Cursor(Cursor.HAND_CURSOR));
            } else {
                setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
            }
        }
    }

    private class SettingsPane extends Panel implements ActionListener {
        private CheckboxGroup player1Group = new CheckboxGroup();
        private CheckboxGroup player2Group = new CheckboxGroup();
        private Checkbox computer1 = new Checkbox("Computer", player1Group, true);
        private Checkbox computer2 = new Checkbox("Computer", player2Group, false);
        private Checkbox human1 = new Checkbox("Human", player1Group, false);
        private Checkbox human2 = new Checkbox("Human", player2Group, true);
        private Button start = new Button("Start Game!");
        private Scrollbar difficulty = new Scrollbar(Scrollbar.HORIZONTAL,
                3, 1, 1, 6);
        private TextField name1 = new TextField("Short");
        private TextField name2 = new TextField("Cut");

	private Font defaultFont = new Font("Dialog", Font.PLAIN, 12);
	private Font largeFont = new Font("Dialog", Font.BOLD, 14);

	public void actionPerformed(ActionEvent e) {
	    GraphGenerator gen = new GraphGenerator();

	    int df = difficulty.getValue();

	    switch (df) {
	    case 1:
		gen.minEdgesForNode = 3; gen.maxEdgesForNode = 3;
                break;
	    case 2:
		gen.minEdgesForNode = 3; gen.maxEdgesForNode = 4;
                break;
	    case 3:
		gen.minEdgesForNode = 3; gen.maxEdgesForNode = 5;
                break;
	    case 4:
		gen.minEdgesForNode = 4; gen.maxEdgesForNode = 6;
                break;
	    case 5:
		gen.minEdgesForNode = 4; gen.maxEdgesForNode = 7;
                break;
	    }

	    field.graph = gen.generateGraph();

	    Player player1, player2;

	    if (human1.getState()) {
		player1 = new HumanPlayer(name1.getText(),
					  Player.SHORT, field);
	    } else {
		player1 = new ComputerPlayer(name1.getText(),
					     Player.SHORT, field);
	    } 

	    if (human2.getState()) {
		player2 = new HumanPlayer(name2.getText(),
					  Player.CUT, field);
	    } else {
		player2 = new ComputerPlayer(name2.getText(),
					     Player.CUT, field);
	    }

	    showField();
	    field.newGame(player1, player2);
	}

	SettingsPane() {
            setBackground(settingsBackground);
	    setFont(defaultFont);

	    computer2.setEnabled(false);

	    start.addActionListener(this);

	    Insets i1 = new Insets(5, 5, 5, 5);
	    Insets i2 = new Insets(5, 5, 0, 5);
	    Insets i3 = new Insets(0, 5, 5, 5);

	    GridBagLayout gbl = new GridBagLayout();
	    setLayout(gbl);
	    GridBagConstraints c = new GridBagConstraints();
	    c.weightx = 1; c.weighty = 1;
	    c.insets = i2;
	    c.anchor = c.CENTER;
	    c.gridwidth = c.REMAINDER;
	    Label title = new Label("Players");
	    title.setFont(largeFont);
	    add(title, c);
	    c.insets = i3;
	    c.anchor = c.WEST;
	    c.gridwidth = c.RELATIVE;
	    c.fill = c.HORIZONTAL;
	    add(name1, c);
	    c.gridwidth = c.REMAINDER;
	    add(name2, c);
	    c.gridwidth = c.RELATIVE;
	    c.insets = i2;
	    c.fill = c.NONE;
	    add(human1, c);
	    c.gridwidth =  c.REMAINDER;
	    add(human2, c);
	    c.gridwidth = c.RELATIVE;
	    c.insets = i3;
	    add(computer1, c);
	    c.gridwidth =  c.REMAINDER;	
	    add(computer2, c);
	    c.gridwidth =  c.REMAINDER;
	    c.anchor = c.CENTER;
	    c.insets = i2;
	    add(new Label("<- Short -- Difficulty -- Cut ->"), c);
	    c.fill = c.HORIZONTAL;
	    c.insets = i3;
	    add(difficulty, c);
	    c.fill = c.NONE;
	    c.anchor = c.CENTER;
	    c.gridwidth = c.REMAINDER;
	    c.insets = i1;
	    start.setFont(largeFont);
	    add(start, c);
	}
    }
}
